#!/usr/bin/env python3

from datetime import timedelta, datetime
from hashlib import sha3_256
from pathlib import Path
import sqlite3
from textwrap import indent
from importlib.resources import files

from typing import Optional


with files(__package__).joinpath("schema.sql").open("r") as fp:
    PRELUDE = fp.read()


def requires_conn(f):
    def _helper(self, *args, **kwargs):
        if self._conn is None:
            raise ConnectionError(f"A connection is required for {f.__name__}")
        return f(self, *args, **kwargs)

    return _helper


def fmap(ctor):
    """Fmap a constructor over the records returned from a query function."""

    def _h1(f):
        def _h2(*args, **kwargs):
            res = f(*args, **kwargs)
            if isinstance(res, list):
                return [ctor(*it) for it in res]
            elif isinstance(res, tuple):
                return ctor(*res)

        return _h2

    return _h1


def one(it, *args, **kwargs):
    assert not args
    assert not kwargs
    return it


class Store(object):
    def __init__(self, path):
        self._path = path
        self._conn: sqlite3.Connection = None

    def connect(self):
        if not self._conn:
            self._conn = sqlite3.connect(self._path, isolation_level="IMMEDIATE")
            for hunk in PRELUDE.split("\n\n"):
                try:
                    self._conn.executescript(hunk).fetchall()
                except sqlite3.OperationalError as e:
                    raise RuntimeError(
                        f"Unable to execute startup script:\n{indent(hunk, ' > ')}"
                    ) from e

    @requires_conn
    def commit(self):
        self._conn.commit()

    def close(self):
        if self._conn:
            self.commit()
            self._conn.close()
            self._conn = None

    @fmap(one)
    @requires_conn
    def try_create_user(self, username, password):
        """Attempt to create a new user."""

        digest = sha3_256()
        digest.update(password.encode("utf-8"))
        return self._conn.execute(
            "INSERT INTO users (name, hash) VALUES (?, ?) RETURNING (id)",
            [username, digest.hexdigest()],
        ).fetchone()

    @requires_conn
    def fetch_user(self, uid: int):
        return self._conn.execute("SELECT * FROM users WHERE id = ?", [uid]).fetchone()

    @fmap(one)
    @requires_conn
    def fetch_user_priority(self, uid: int):
        return self._conn.execute(
            "SELECT priority FROM groups g INNER JOIN users u ON g.id = u.group_id WHERE u.id = ?",
            [uid],
        ).fetchone()

    @requires_conn
    def list_users(self):
        return self._conn.execute("SELECT id, name FROM users").fetchall()

    ################################################################################
    # Sessions / 'keys'

    @fmap(one)
    @requires_conn
    def _create_session(self, uid: int, ttl: Optional[timedelta]):
        return self._conn.execute(
            "INSERT INTO user_keys (user_id, expiration) VALUES (?, ?) RETURNING (id)",
            [uid, (datetime.now() + ttl).isoformat() if ttl else None],
        ).fetchone()

    @requires_conn
    def try_login(self, username: str, password: str, ttl: timedelta) -> Optional[str]:
        """Given a username and an (unsecured) password, attempt to authenticate the named user.

        If successful, return the ID of a new session/key for that user.

        """

        digest = sha3_256()
        digest.update(password.encode("utf-8"))
        res = self._conn.execute(
            "SELECT id FROM users WHERE name=? AND hash=? LIMIT 1",
            [username, digest.hexdigest()],
        ).fetchone()
        if not res:
            return
        uid = res[0]
        return self._create_session(uid, ttl)

    @requires_conn
    def create_key(self, kid: str, ttl) -> Optional[str]:
        """Given an _existing_ login session, create a new key.

        This allows the user to create more or less permanent API keys associated with their identity.

        """

        if uid := self.try_key(kid):
            return self._create_session(uid, ttl)

    @requires_conn
    def list_keys(self):
        return self._conn.execute("SELECT id, user_id FROM user_keys").fetchall()

    @requires_conn
    def fetch_key(self, kid) -> tuple:
        return self._conn.execute(
            "SELECT * FROM user_keys WHERE id = ?", [kid]
        ).fetchone()

    @fmap(one)
    @requires_conn
    def try_key(self, kid: str):
        """Try to find the mapped user for a session."""

        return self._conn.execute(
            "SELECT user_id FROM user_keys WHERE expiration IS NULL OR unixepoch(expiration) > unixepoch('now') and id = ?",
            [kid],
        ).fetchone()

    @requires_conn
    def refresh_key(self, kid: str, ttl: timedelta):
        """Automagically renew an API key which is still in use.

        Mostly intended for dealing with web sessions which should implicitly extend, but which use the same mechanism as API keys.

        """

        self._conn.execute(
            "UPDATE user_keys SET expiration = ? WHERE id = ?",
            [(datetime.now() + ttl).isoformat(), kid],
        )

    @requires_conn
    def delete_key(self, kid: str):
        """Remove a session/key; equivalent to logout."""

        self._conn.execute("DELETE FROM user_keys WHERE id = ?", [kid])

    ################################################################################
    # Printers
    #
    # Printers represent connections to OctoPrint instances controlling physical machines.
    @requires_conn
    def create_printer(self):
        pass

    @requires_conn
    def list_printers(self):
        pass

    @requires_conn
    def update_printer_status(self):
        pass

    ################################################################################
    # Files
    #
    # A record of local files on disk, and the users who own then.
    @fmap(one)
    @requires_conn
    def create_file(self, uid: int, name: str, path: Path) -> int:
        return self._conn.execute(
            "INSERT INTO files (user_id, filename, upload_date) VALUES (?, ?, datetime('now')) RETURNING (id)",
            [uid, name],
        ).fetchone()

    @requires_conn
    def list_files(self, uid: int):
        return self._conn.execute(
            "SELECT * FROM files WHERE user_id = ?", [uid]
        ).fetchall()

    @requires_conn
    def delete_file(self, uid: int, fid: int):
        self._conn.execute("DELETE FROM files WHERE user_id = ? AND id = ?", [uid, fid])

    ################################################################################
    # Job
    #
    # A request by a user for a given file to be printed.
    @fmap(one)
    @requires_conn
    def create_job(self, uid: int, fid: int, relative_priority: int = 0):
        """Create a job mapped to a user with a file to print and a priority.

        Note that the user may provide a sub-priority within their group queue. This allows users to create jobs with
        higher priority than existing jobs as a means of controlling the queue order.

        May want a different (eg. more explicit) queue ordering mechanism here. Emulating the Netflix queue behavior of
        being able to drag some jobs ahead of others? How to model that?

        """
        assert 0 <= relative_priority <= 9
        return self._conn.execute(
            "INSERT INTO jobs (user_id, file_id, priority) VALUES (?, ?, ?) RETURNING (id)",
            [uid, fid, self.fetch_user_priority(uid) + relative_priority],
        ).fetchone()

    @requires_conn
    def list_jobs(self, uid: Optional[int] = None):
        """Enumerate jobs in priority order."""
        cond = f"AND user_id = {uid}" if uid else ""
        return self._conn.execute(
            f"SELECT * FROM jobs WHERE started_at IS NULL {cond} ORDER BY priority DESC",
            [],
        ).fetchall()

    @requires_conn
    def fetch_job(self, uid: int, jid: int) -> Optional[tuple]:
        return self._conn.execute(
            "SELECT * FROM jobs WHERE user_id = ? AND id = ?", [uid, jid]
        ).fetchone()

    @requires_conn
    def alter_job(self, job: tuple):
        fields = [
            "id",
            "user_id",
            "file_id",
            "priority",
            "started_at",
            "finished_at",
            "state",
            "printer_id",
        ]
        assert len(job) == len(fields)
        return self._conn.execute(
            f"INSERT OR REPLACE INTO jobs ({', '.join(fields)}) VALUES ({', '.join(['?'] * len(fields))})",
            job,
        )

    @requires_conn
    def delete_job(self, uid: int, jid: int):
        self._conn.execute("DELETE FROM jobs WHERE user_id = ? and id = ?", [uid, jid])
